from typing import List, Union, Tuple, Callable, Any
from .expression_parser import ExpressionParser
from .sql_parser import SqlParser
from .types import TermTuple


def expression(domain: List[Union[str, TermTuple]]) -> Callable:
    parser = ExpressionParser()
    return parser.parse(domain)


def sql(domain: List[Union[str, TermTuple]]) -> Tuple[Any, Any]:
    parser = SqlParser()
    return parser.parse(domain)



import operator
from typing import List, Union, Callable, Any, Dict, Tuple
from types import SimpleNamespace
from fnmatch import fnmatchcase
from .types import TermTuple


class ExpressionParser:

    def __init__(self, evaluator: Callable = lambda x: x) -> None:
        self.evaluator = evaluator

        self.comparison_dict = {
            '=': operator.eq,
            '!=': operator.ne,
            '<=': operator.le,
            '<': operator.lt,
            '>': operator.gt,
            '>=': operator.ge,
            'in': lambda x, y: x in y,
            'like': lambda x, y: self._parse_like(x, y),
            'ilike': lambda x, y: self._parse_like(x, y, True),
            'contains': operator.contains
        }

        self.binary_dict = {
            '&': lambda expression_1, expression_2: (
                lambda obj: (expression_1(obj) and expression_2(obj))),
            '|': lambda expression_1, expression_2: (
                lambda obj: (expression_1(obj) or expression_2(obj)))
        }

        self.unary_dict = {
            '!': lambda expression_1: (
                lambda obj: (not expression_1(obj)))
        }

        self.default_join_operator = '&'

    def parse(self, domain: List[Union[str, TermTuple]],
              namespaces: List[str] = []) -> Callable:
        if not domain:
            return lambda obj: True
        stack: List[Callable] = []
        for item in list(reversed(domain)):
            if isinstance(item, str) and item in self.binary_dict:
                first_operand = stack.pop()
                second_operand = stack.pop()
                function = self.binary_dict[str(item)](
                    first_operand, second_operand)
                stack.append(function)
            elif isinstance(item, str) and item in self.unary_dict:
                operand = stack.pop()
                stack.append(self.unary_dict[str(item)](operand))

            stack = self._default_join(stack)

            if isinstance(item, (list, tuple)):
                result = self._parse_term(item, namespaces)
                stack.append(result)

        result = self._default_join(stack)[0]
        return result

    def _default_join(self, stack: List[Callable]) -> List[Callable]:
        operator = self.default_join_operator
        if len(stack) == 2:
            first_operand = stack.pop()
            second_operand = stack.pop()
            function = self.binary_dict[operator](
                first_operand, second_operand)
            stack.append(function)
        return stack

    def _parse_term(self, term_tuple: TermTuple,
                    namespaces: List[str] = []) -> Callable:
        field, operator, value = term_tuple
        value = self.evaluator(value)
        comparator = self.comparison_dict.get(operator)
        return self._build_filter(field, comparator, value, namespaces)

    def _build_filter(self, field, comparator, value, namespaces=[]):
        def function(obj):
            obj_, field_, value_ = self._process_namespaces(
                obj, field, value, namespaces)
            return comparator(getattr(obj_, field_), value_)
        return function

    def _process_namespaces(self, obj, field, value, namespaces):
        if not namespaces or not isinstance(obj, tuple):
            if isinstance(obj, dict):
                obj = SimpleNamespace(**obj)
            return obj, field, value

        base_object = (
            SimpleNamespace(**obj[0]) if
            isinstance(obj[0], dict) else obj[0])

        for i, namespace in enumerate(namespaces):
            namespace_object = (
                SimpleNamespace(**obj[i]) if
                isinstance(obj[i], dict) else obj[i])

            if value.startswith(f"{namespace}."):
                _, attribute = value.split('.')
                value = getattr(namespace_object, attribute)

            if field.startswith(f"{namespace}."):
                _, field = field.split('.')
                base_object = namespace_object

        return base_object, field, value

    @staticmethod
    def _parse_like(value: str, pattern: str, insensitive=False) -> bool:
        if not isinstance(value, str):
            return False
        pattern = pattern.replace('%', '*').replace('_', '?')
        pattern = pattern.lower() if insensitive else pattern
        value = value.lower() if insensitive else value
        return fnmatchcase(value, pattern)



import calendar
import datetime
import time
from typing import Dict, Any


class SafeEval:

    def __init__(self) -> None:
        self._prefix = '>>>'
        self._globals = {
            '__builtins__': None
        }
        self._forbidden = ['__', '**']
        self._safe_builtins = {
            'abs': abs,
            'calendar': calendar,
            'datetime': datetime,
            'float': float,
            'int': int,
            'max': max,
            'min': min,
            'round': round,
            'str': str,
            'sum': sum,
            'time': time
        }

    def __call__(self, expression: str, locals: Dict[str, Any] = None) -> Any:
        if not isinstance(expression, str) or not expression.startswith(
                self._prefix) or any(
                    character in expression for character in self._forbidden):
            return expression

        if not locals:
            locals = {}

        expression = expression.replace(self._prefix, '').strip()
        locals.update(self._safe_builtins)

        return eval(expression, self._globals, locals)



import unittest
import filtrark

from unittest.mock import Mock


class TestApi(unittest.TestCase):

    def setUp(self):
        self.domain = [('field', '=', 5), ('field2', '=', 4)]
        self.mock_object = Mock(field=5, field2=4)

    def test_api_expression(self):
        function = filtrark.expression(self.domain)

        def expected(obj):
            return obj.field == 5 and obj.field2 == 4

        self.assertEqual(function(self.mock_object),
                         expected(self.mock_object))

    def test_api_sql(self):
        result = filtrark.sql(self.domain)
        expected = ('field = %s AND field2 = %s', (5, 4))
        self.assertEqual(result, expected)



import unittest
import datetime
from filtrark.safe_eval import SafeEval


class TestSafeEval(unittest.TestCase):

    def setUp(self):
        self.safe_eval = SafeEval()

    def test_safe_eval_object_creation(self):
        self.assertTrue(isinstance(self.safe_eval, SafeEval))

    def test_safe_eval_doesnt_have_globals_or_locals_by_default(self):
        with self.assertRaises(TypeError) as e:
            self.safe_eval('>>> dir()')

    def test_safe_eval_doesnt_evaluate_unprefixed_strings(self):
        result = self.safe_eval('8 + 8')
        self.assertEqual(result, '8 + 8')

    def test_safe_eval_evaluate_simple_expression(self):
        result = self.safe_eval('>>> 2 + 2')
        self.assertEqual(result, 4)

    def test_safe_eval_forbidden_characters(self):
        forbidden_expressions = [
            '>>> __name__',
            '>>> 4 ** 4'
        ]

        for expression in forbidden_expressions:
            result = self.safe_eval(expression)
            self.assertEqual(result, expression)

    def test_safe_eval_safe_builtins(self):
        forbidden_expressions = [
            ('>>> datetime.datetime.now()', datetime.datetime),
            ('>>> max(1, 2)', int),
            ('>>> float(5)', float),
            ('>>> int(time.time()) + 3600', int)
        ]

        for expression, expected in forbidden_expressions:
            result = self.safe_eval(expression)
            self.assertEqual(type(result), expected)

    def test_safe_eval_custom_locals(self):
        expression = '>>> custom_function()'
        custom_locals = {
            'custom_function': lambda: 77
        }

        result = self.safe_eval(expression, custom_locals)
        self.assertEqual(result, 77)

    def test_safe_eval_only_works_with_strings(self):
        expression = True
        result = self.safe_eval(expression)
        self.assertEqual(result, expression)

        expression = 55
        result = self.safe_eval(expression)
        self.assertEqual(result, expression)



import unittest
from fnmatch import fnmatchcase
from unittest.mock import Mock
from filtrark.safe_eval import SafeEval
from filtrark.expression_parser import ExpressionParser


class TestExpressionParser(unittest.TestCase):

    def setUp(self):
        self.parser = ExpressionParser()

    def test_expression_parser_object_creation(self):
        self.assertTrue(isinstance(self.parser, ExpressionParser))

    def test_expression_parser_parse_tuple(self):
        filter_tuple_list = [
            (('field', '=', 9), lambda obj: obj.field == 9, Mock(field=9)),
            (('field', '!=', 9), lambda obj: obj.field != 9, Mock(field=8)),
            (('field', '>', 9), lambda obj: obj.field > 9, Mock(field=10)),
            (('field', '<', 9), lambda obj: obj.field < 9, Mock(field=8)),
            (('field', '<=', 9), lambda obj: obj.field <= 9, Mock(field=8)),
            (('field', '>=', 9), lambda obj: obj.field >= 9, Mock(field=10)),
            (('field', 'in', [1, 2, 3]), lambda obj: obj.field in [1, 2, 3],
             Mock(field=2))
        ]

        for test_tuple in filter_tuple_list:
            filter_tuple = test_tuple[0]
            expected_function = test_tuple[1]
            mock_object = test_tuple[2]
            function = self.parser._parse_term(filter_tuple)

            self.assertTrue(callable(function))
            self.assertTrue(function(mock_object))
            self.assertEqual(
                function(mock_object), expected_function(mock_object))

    def test_expression_parser_parse_single_term(self):
        domain = [('field', '=', 7)]
        expected = 'field = 7'

        def expected(obj):
            return getattr(obj, 'field') == 7

        mock_object = Mock()
        mock_object.field = 7

        function = self.parser.parse(domain)

        self.assertTrue(callable(function))
        self.assertTrue(function(mock_object))
        self.assertEqual(function(mock_object), expected(mock_object))

        mock_object.field = 5
        self.assertFalse(function(mock_object))

    def test_expression_parser_default_join(self):
        stack = [lambda obj: obj.field2 != 8, lambda obj: obj.field == 7]

        def expected(obj):
            return (obj.field == 7 and obj.field2 != 8)

        result_stack = self.parser._default_join(stack)

        mock_object = Mock()
        mock_object.field = 7
        mock_object.field2 = 9

        self.assertTrue(isinstance(result_stack, list))
        self.assertEqual(result_stack[0](mock_object), expected(mock_object))
        self.assertTrue(result_stack[0](mock_object))

        mock_object.field = 5
        self.assertFalse(result_stack[0](mock_object))

    def test_expression_parser_parse_multiple_terms(self):
        test_domains = [
            ([('field', '=', 7), ('field2', '!=', 8)],
             lambda obj: (obj.field2 != 8 and obj.field == 7),
             Mock(field=7, field2=8)),
            ([('field', '=', 7), ('field2', '!=', 8), ('field3', '>=', 9)],
             (lambda obj: (obj.field2 != 8 and
                           obj.field == 7 and obj.field3 >= 9)),
             Mock(field=7, field2=5, field3=9)),
            (['|', ('field', '=', 7), ('field2', '!=', 8)],
             lambda obj: (obj.field2 != 8 or obj.field == 7),
             Mock(field=7, field2=8)),
            (['|', ('field', '=', 7), '!', ('field2', '!=', 8),
              ('field3', '>=', 9)],
                (lambda obj: (obj.field == 7 or
                              not obj.field2 != 8 and obj.field3 >= 9)),
             Mock(field=7, field2=8, field3=9)),
            (['!', ('field', '=', 7)],
             lambda obj: (not obj.field == 7),
             Mock(field=7)),
        ]

        for test_domain in test_domains:
            result = self.parser.parse(test_domain[0])
            expected = test_domain[1]
            obj = test_domain[2]
            self.assertTrue(callable(result))
            self.assertEqual(result(obj), expected(obj))

    def test_expression_parser_with_empty_list(self):
        domain = []
        result = self.parser.parse(domain)
        mock_object = Mock()
        mock_object.field = 7
        self.assertTrue(result(mock_object))

    def test_expression_parser_with_lists_of_lists(self):
        domain = [['field', '=', 7], ['field2', '!=', 8]]

        def expected(obj):
            return (obj.field == 7 and obj.field2 != 8)

        result = self.parser.parse(domain)

        mock_object = Mock()
        mock_object.field = 7
        mock_object.field2 = 9

        self.assertTrue(result(mock_object))
        self.assertEqual(result(mock_object), expected(mock_object))

    def test_expression_parser_parse_like(self):
        filter_tuple_list = [
            (('field', 'like', "Hello"),
             lambda obj: fnmatchcase(obj.field, "Hello"),
             Mock(field="Hello")),
            (('field', 'like', "Hello%"),
             lambda obj: fnmatchcase(obj.field, "Hello*"),
             Mock(field="Hello World")),
            (('field', 'like', "%World"),
             lambda obj: fnmatchcase(obj.field, "*World"),
             Mock(field="Hello World")),
            (('field', 'like', "Hello_"),
             lambda obj: fnmatchcase(obj.field, "Hello?"),
             Mock(field="HelloX")),
            (('field', 'ilike', "%World"),
             lambda obj: fnmatchcase(obj.field, "*world"),
             Mock(field="hello world")),
            (('field', 'ilike', "%eLLo%"),
             lambda obj: fnmatchcase(obj.field, "*ello*"),
             Mock(field="hello world")),
            (('field', 'like', "Hello%"),
             lambda obj: False, Mock(field=9)),
        ]

        for test_tuple in filter_tuple_list:
            filter_tuple = test_tuple[0]
            expected_function = test_tuple[1]
            mock_object = test_tuple[2]

            function = self.parser._parse_term(filter_tuple)

            self.assertTrue(callable(function))
            self.assertEqual(function(mock_object),
                             expected_function(mock_object))

    def test_expression_parser_parse_contains(self):
        filter_tuple_list = [
            (('field', 'contains', "333"), lambda obj: '333' in obj.field,
             Mock(field=['1', '22', '333']))]

        for test_tuple in filter_tuple_list:
            filter_tuple = test_tuple[0]
            expected_function = test_tuple[1]
            mock_object = test_tuple[2]

            function = self.parser._parse_term(filter_tuple)

            self.assertTrue(callable(function))
            self.assertEqual(
                function(mock_object), expected_function(mock_object))

    def test_expression_parser_parse_evaluator(self):
        self.parser.evaluator = SafeEval()
        domain = [('field', '=', '>>> 3 + 4')]
        result = self.parser.parse(domain)
        mock_object = Mock(field=7)
        self.assertTrue(result(mock_object))

    def test_expression_parser_namespaces(self):
        namespaces = ['orders', 'customers']
        domain = [('orders.customer_id', '=', 'customers.id')]

        result = self.parser.parse(domain, namespaces)
        mock_object = (
            {'id': '77', 'customer_id': '03'}, {'id': '03', 'name': 'Joe'})

        result(mock_object)

        self.assertTrue(result(mock_object))



from typing import List, Union, Tuple, Any, Callable
from .types import TermTuple


class SqlParser:

    def __init__(self, evaluator: Callable = lambda x: x) -> None:
        self.evaluator = evaluator

        self.comparison_dict = {
            '=': lambda x, y:  ' = '.join([str(x), str(y)]),
            '!=': lambda x, y: ' <> '.join([str(x), str(y)]),
            '<=': lambda x, y: ' <= '.join([str(x), str(y)]),
            '<': lambda x, y: ' < '.join([str(x), str(y)]),
            '>': lambda x, y: ' > '.join([str(x), str(y)]),
            '>=': lambda x, y: ' >= '.join([str(x), str(y)]),
            'in': lambda x, y: ' IN '.join([str(x), str(y)]),
            'like': lambda x, y: "{0} LIKE {1}".format(str(x), str(y)),
            'ilike': lambda x, y: "{0} ILIKE {1}".format(str(x), str(y)),
            'contains': lambda x, y: '{0} @> {{{1}}}'.format(str(x), str(y))
        }

        self.binary_dict = {
            '&': lambda a, b: a + ' AND ' + b,
            '|': lambda a, b: a + ' OR ' + b}

        self.unary_dict = {
            '!': lambda a: 'NOT ' + a}

        self.default_join_operator = '&'

    def parse(self, domain: List[Union[str, TermTuple]],
              namespaces: List[str] = []) -> Tuple:
        if not domain:
            return "TRUE", ()
        stack = []  # type: List[str]
        params = []
        for item in list(reversed(domain)):
            if isinstance(item, str) and item in self.binary_dict:
                first_operand = stack.pop()
                second_operand = stack.pop()
                string_term = str(
                    self.binary_dict[str(item)](
                        first_operand, second_operand))
                stack.append(string_term)
            elif isinstance(item, str) and item in self.unary_dict:
                operand = stack.pop()
                stack.append(
                    str(self.unary_dict[str(item)](
                        operand)))

            stack = self._default_join(stack)

            if isinstance(item, (list, tuple)):
                result_tuple = self._parse_term(item)
                stack.append(result_tuple[0])
                params.append(result_tuple[1])

        result_query = str(self._default_join(stack)[0])
        result = [result_query, tuple(reversed(params))]

        if namespaces:
            result.append(", ".join(namespaces))

        return tuple(result)

    def _default_join(self, stack: List[str]) -> List[str]:
        operator = self.default_join_operator
        if len(stack) == 2:
            first_operand = stack.pop()
            second_operand = stack.pop()
            value = str(self.binary_dict[operator](
                str(first_operand), str(second_operand)))
            stack.append(value)
        return stack

    def _parse_term(self, term_tuple: TermTuple) -> Tuple[str, Any]:
        field, operator, value = term_tuple
        if isinstance(value, list):
            value = tuple(value)
        if isinstance(value, str):
            value = self.evaluator(value)
        function = self.comparison_dict.get(operator)
        placeholder = '%s'
        result = (function(field, placeholder), value)
        return result



import unittest
from unittest.mock import Mock
from filtrark.safe_eval import SafeEval
from filtrark.sql_parser import SqlParser


class TestSqlParser(unittest.TestCase):

    def setUp(self):
        self.parser = SqlParser()

    def test_sql_parser_object_creation(self):
        self.assertTrue(isinstance(self.parser, SqlParser))

    def test_sql_parser_parse_tuple(self):
        filter_tuple_list = [
            (('field', '=', 99), ('field = %s', 99)),
            (('field', 'like', 'world'), ("field LIKE %s", "world")),
            (('field', 'ilike', 'world'), ("field ILIKE %s", "world")),
            (('field', 'contains', 99), ("field @> {%s}", 99))
        ]
        for filter_tuple, expected in filter_tuple_list:
            result = self.parser._parse_term(filter_tuple)
            self.assertEqual(result, expected)

    def test_sql_parser_parse_single_term(self):
        domain = [('field', '=', 7)]
        expected = ('field = %s', (7,))
        result = self.parser.parse(domain)
        self.assertEqual(result, expected)

    def test_sql_parser_default_join(self):
        stack = ['field2 <> %s', 'field = %s']
        expected = 'field = %s AND field2 <> %s'
        result = self.parser._default_join(stack)
        self.assertEqual(result, [expected])

    def test_string_parser_parse_multiple_terms(self):
        test_domains = [
            ([('field', '=', 7), ('field2', '!=', 8)],
             ('field = %s AND field2 <> %s', (7, 8))),
            ([('field', '=', 7), ('field2', '!=', 8), ('field3', '>=', 9)],
             ('field = %s AND field2 <> %s AND field3 >= %s', (7, 8, 9))),
            (['|', ('field', '=', 7), ('field2', '!=', 8),
              ('field3', '>=', 9)],
                ('field = %s OR field2 <> %s AND field3 >= %s', (7, 8, 9))),
            (['|', ('field', '=', 7),
              '!', ('field2', '!=', 8), ('field3', '>=', 9)],
             ('field = %s OR NOT field2 <> %s AND field3 >= %s', (7, 8, 9))),
            (['!', ('field', '=', 7)], ('NOT field = %s', (7,))),
        ]

        for test_domain in test_domains:
            result = self.parser.parse(test_domain[0])
            expected = test_domain[1]
            self.assertEqual(result, expected)

    def test_sql_parser_with_empty_list(self):
        domain = []
        result = self.parser.parse(domain)
        expected = "TRUE", ()
        self.assertEqual(result, expected)

    def test_sql_parser_with_lists_of_lists(self):
        domain = [['field', '=', 7], ['field2', '!=', 8]]
        expected = ('field = %s AND field2 <> %s', (7, 8))
        result = self.parser.parse(domain)
        self.assertEqual(result, expected)

    def test_sql_parser_with_lists_parameters(self):
        domain = [['field', 'in', [7]]]
        expected = ('field IN %s', ((7,),))
        result = self.parser.parse(domain)
        self.assertEqual(result, expected)

    def test_sql_parser_parse_evaluator(self):
        self.parser.evaluator = SafeEval()
        domain = [('field', '=', '>>> 3 + 4')]
        expected = ('field = %s', (7,))
        result = self.parser.parse(domain)
        self.assertEqual(result, expected)

    def test_sql_parser_namespaces(self):
        namespaces = ['orders', 'customers']
        domain = [('orders.customer_id', '=', 'customers.id')]

        expected = ('orders.customer_id = %s',
                    ('customers.id',), 'orders, customers')

        result = self.parser.parse(domain, namespaces)

        self.assertEqual(result, expected)



# -*- coding: utf-8 -*-

"""Top-level package for Filtrark."""

__author__ = """Esteban Echeverry"""
__email__ = 'eecheverry@nubark.com'
__version__ = '0.6.2'


from .api import expression, sql
from .expression_parser import ExpressionParser
from .safe_eval import SafeEval
from .sql_parser import SqlParser
